const std = @import("std");

pub const Token = union(enum) {
    invalid: []const u8,
    literal: []const u8,
    escaped: u8,
    placeholder: []const u8,
};

pub const TokenStream = struct {
    format: []const u8,

    pub fn chop(self: *TokenStream, len: usize) []const u8 {
        const token = self.format[0..len];

        self.skip(len);

        return token;
    }

    pub fn init(format: []const u8) TokenStream {
        return .{
            .format = format,
        };
    }

    pub fn next(self: *TokenStream) ?Token {
        if (self.format.len != 0) {
            switch (self.format[0]) {
                '}' => {
                    if (self.peek(1) != '}') {
                        return .{ .invalid = self.chop(1) };
                    }

                    self.skip(1);

                    return .{ .escaped = self.chop(1)[0] };
                },

                '{' => {
                    self.skip(1);

                    if (self.format.len != 0) {
                        if (self.format[0] == '{') {
                            return .{ .escaped = self.chop(1)[0] };
                        }

                        if (std.mem.indexOfScalar(u8, self.format, '}')) |end_index| {
                            defer {
                                self.skip(1);
                            }

                            return .{ .placeholder = self.chop(end_index) };
                        }
                    }

                    return .{ .invalid = self.chop(self.format.len) };
                },

                else => {
                    if (std.mem.indexOfScalar(u8, self.format, '{')) |placeholder_index| {
                        return .{ .literal = self.chop(placeholder_index) };
                    }

                    const end_index = std.mem.indexOfScalar(u8, self.format, '}') orelse self.format.len;

                    return .{ .literal = self.chop(end_index) };
                },
            }
        }

        return null;
    }

    pub fn peek(self: TokenStream, ahead_index: usize) ?u8 {
        if (self.format.len < (ahead_index + 1)) {
            return null;
        }

        return self.format[ahead_index];
    }

    pub fn skip(self: *TokenStream, len: usize) void {
        self.format = self.format[len..];
    }
};

test "complex sequence" {
    var stream = TokenStream.init("layout (binding = {binding}) uniform {identifier} {{ }}\n");

    try std.testing.expectEqualStrings("layout (binding = ", stream.next().?.literal);
    try std.testing.expectEqualStrings("binding", stream.next().?.placeholder);
    try std.testing.expectEqualStrings(") uniform ", stream.next().?.literal);
    try std.testing.expectEqualStrings("identifier", stream.next().?.placeholder);
    try std.testing.expectEqualStrings(" ", stream.next().?.literal);
    try std.testing.expectEqual('{', stream.next().?.escaped);
    try std.testing.expectEqualStrings(" ", stream.next().?.literal);
    try std.testing.expectEqual('}', stream.next().?.escaped);
    try std.testing.expectEqualStrings("\n", stream.next().?.literal);
    try std.testing.expectEqual(null, stream.next());
}

test "invalid placeholder symbol usage" {
    var stream = TokenStream.init("} {identifier};\n");

    try std.testing.expectEqualStrings("}", stream.next().?.invalid);
    try std.testing.expectEqualStrings(" ", stream.next().?.literal);
    try std.testing.expectEqualStrings("identifier", stream.next().?.placeholder);
    try std.testing.expectEqualStrings(";\n", stream.next().?.literal);
}
