pub const Box = @import("./Box.zig");

pub const TypeId = @import("./TypeId.zig");

pub const asio = @import("./asio.zig");

pub const bytes = @import("./bytes.zig");

pub const hashes = @import("./hashes.zig");

pub const heap = @import("./heap.zig");

pub const map = @import("./map.zig");

pub const scalars = @import("./scalars.zig");

pub const shaders = @import("./shaders.zig");

const std = @import("std");

pub const utf8 = @import("./utf8.zig");

pub fn Callable(comptime Output: type, comptime input_types: []const type) type {
    const InputTuple = std.meta.Tuple(input_types);
    const fn_context: *allowzero anyopaque = @ptrFromInt(0);

    return struct {
        context: *allowzero anyopaque,
        call_with_context: *const fn (*anyopaque, InputTuple) Output,

        fn FnCall(comptime invoke: anytype) type {
            return struct {
                fn call(context: *anyopaque, inputs: InputTuple) Output {
                    std.debug.assert(context == fn_context);

                    return @call(.auto, invoke, inputs);
                }
            };
        }

        fn RefCall(comptime ContextPtr: type, comptime invoke: anytype) type {
            const Invoke = @TypeOf(invoke);
            const is_zero_aligned = @typeInfo(ContextPtr).pointer.alignment == 0;

            return struct {
                fn invoke_concrete(context: *anyopaque, inputs: InputTuple) Output {
                    std.debug.assert(context != fn_context);

                    if (is_zero_aligned) {
                        return @call(.auto, invoke, .{@as(ContextPtr, @ptrCast(context))} ++ inputs);
                    }

                    var args: std.meta.ArgsTuple(Invoke) = undefined;

                    args[0] = @as(ContextPtr, @ptrCast(@alignCast(context)));

                    inline for (1..args.len, &inputs) |i, input| {
                        args[i] = input;
                    }

                    return switch (@typeInfo(Output)) {
                        .error_union => try @call(.auto, invoke, args),
                        else => @call(.auto, invoke, args),
                    };
                }
            };
        }

        const Self = @This();

        pub fn call(self: Self, inputs: InputTuple) Output {
            return self.call_with_context(self.context, inputs);
        }

        pub fn initFn(comptime invoke: anytype) Self {
            return .{
                .context = fn_context,
                .call_with_context = FnCall(invoke).call,
            };
        }

        pub fn initRef(context_ptr: anytype, comptime invoke: anytype) Self {
            return .{
                .context = @ptrCast(context_ptr),
                .call_with_context = RefCall(@TypeOf(context_ptr), invoke).invoke_concrete,
            };
        }

        pub fn is_fn(self: Self, invoke: anytype) bool {
            const Invoke = @TypeOf(invoke);

            return switch (@typeInfo(Invoke)) {
                .pointer => self.is_fn(invoke.*),
                .@"fn" => (self.context == fn_context) and (self.call_with_context == FnCall(invoke).call),
                else => @compileError(std.fmt.comptimePrint("parameter `fn` must be a function type or pointer to one, not {s}", .{@typeName(Invoke)})),
            };
        }
    };
}

pub const ShortString = extern struct {
    buffer: [max]u8,
    remaining: u8,

    pub const Error = error{
        StringTooLong,
    };

    pub fn append(self: *ShortString, c: u8) Error!void {
        const post_remaining = scalars.sub(self.remaining, 1) orelse {
            return error.StringTooLong;
        };

        self.buffer[self.len()] = c;
        self.remaining = post_remaining;
    }

    pub fn assign(self: *ShortString, data: []const u8) Error!void {
        self.clear();

        try self.join(data);
    }

    pub fn clear(self: *ShortString) void {
        self.* = empty;
    }

    pub fn init(data: []const u8) Error!ShortString {
        var string = empty;

        try string.join(data);

        return string;
    }

    pub fn isEmpty(self: ShortString) bool {
        return self.remaining == max;
    }

    pub fn join(self: *ShortString, data: []const u8) Error!void {
        const remaining = scalars.sub(self.remaining, data.len) orelse {
            return error.StringTooLong;
        };

        @memcpy(self.buffer[self.len()..(max - remaining)], data);

        self.remaining = @intCast(remaining);
    }

    pub const empty = ShortString{
        .buffer = [_]u8{0} ** max,
        .remaining = max,
    };

    pub fn len(self: ShortString) usize {
        return max - self.remaining;
    }

    pub const max = 255;

    pub fn ptr(self: *const ShortString) [*:0]const u8 {
        return @ptrCast(&self.buffer);
    }

    pub fn slice(self: *const ShortString) [:0]const u8 {
        return @ptrCast(self.ptr()[0..self.len()]);
    }

    pub fn writeFormat(self: ShortString, output: bytes.Writable) bytes.ReadWriteError!void {
        return bytes.writeAll(output, self.slice());
    }

    pub fn writer(self: *ShortString) bytes.Writable {
        const writing = struct {
            fn write(string: *ShortString, buffer: []const u8) usize {
                string.append(buffer) catch {
                    return 0;
                };

                return buffer.len;
            }
        };

        return .init(ShortString, self, writing.write);
    }
};

pub fn Stack(comptime Value: type) type {
    return struct {
        values: []Value,
        cap: usize,

        const Self = @This();

        pub fn clear(self: *Self) void {
            self.values = self.values[0..0];
        }

        pub fn deinit(self: *Self) void {
            if (self.cap == 0) {
                return;
            }

            heap.allocator.free(self.values.ptr[0..self.cap]);

            self.* = undefined;
        }

        pub const empty = Self{
            .values = &.{},
            .cap = 0,
        };

        pub fn grow(self: *Self, additional: usize) std.mem.Allocator.Error!void {
            const grown_capacity = self.values.len + additional;

            if (grown_capacity <= self.cap) {
                return;
            }

            const buffer = try heap.allocator.alloc(Value, grown_capacity);

            errdefer {
                heap.allocator.deallocate(buffer);
            }

            if (self.cap != 0) {
                @memcpy(buffer[0..self.values.len], self.values);
                heap.allocator.free(self.values.ptr[0..self.cap]);
            }

            self.values = buffer[0..self.values.len];
            self.cap = grown_capacity;
        }

        pub fn isEmpty(self: Self) bool {
            return self.values.len == 0;
        }

        pub fn isFull(self: Self) bool {
            return self.values.len == self.cap;
        }

        pub fn get(self: Self) ?*Value {
            if (self.values.len == 0) {
                return null;
            }

            return &self.values[self.values.len - 1];
        }

        pub fn pop(self: *Self) ?Value {
            if (self.values.len == 0) {
                return null;
            }

            const tail_index = self.values.len - 1;

            defer self.values = self.values[0..tail_index];

            return self.values[tail_index];
        }

        pub fn push(self: *Self, value: Value) bool {
            if (self.isFull()) {
                return false;
            }

            const offset_index = self.values.len;

            self.values = self.values.ptr[0 .. self.values.len + 1];
            self.values[offset_index] = value;

            return true;
        }

        pub fn pushAll(self: *Self, values: []const Value) bool {
            const new_length = self.values.len + values.len;

            if (new_length > self.cap) {
                return false;
            }

            const offset_index = self.values.len;

            self.values = self.values.ptr[0..new_length];

            for (0..values.len) |index| {
                self.values[offset_index + index] = values[index];
            }

            return true;
        }

        pub fn pushGrow(self: *Self, value: Value) std.mem.Allocator.Error!void {
            try self.grow(@max(1, self.values.len));

            std.debug.assert(self.push(value));
        }

        pub fn pushMany(self: *Self, n: usize, value: Value) bool {
            const new_length = self.values.len + n;

            if (new_length > self.cap) {
                return false;
            }

            const offset_index = self.values.len;

            self.values = self.values.ptr[0..new_length];

            for (0..n) |index| {
                self.values[offset_index + index] = value;
            }

            return true;
        }

        pub fn resize(self: *Self, size: usize, default_value: Value) std.mem.Allocator.Error!void {
            if (self.cap == size) {
                return;
            }

            const values = try heap.allocator.alloc(Value, size);

            for (0..@min(values.len, self.values.len)) |i| {
                values[i] = self.values[i];
            }

            if (values.len > self.values.len) {
                for (self.values.len..values.len) |i| {
                    values[i] = default_value;
                }
            }

            self.values = values[0..values.len];
            self.cap = values.len;
        }
    };
}

pub fn require(function: anytype, args: std.meta.ArgsTuple(@TypeOf(function))) switch (@typeInfo(@typeInfo(@TypeOf(function)).@"fn".return_type.?)) {
    .error_union => |error_union| error_union.payload,
    else => @compileError("fn parameter `function` must return an error union"),
} {
    return @call(.auto, function, args) catch |call_error| {
        if (@errorReturnTrace()) |trace| {
            std.debug.dumpStackTrace(trace.*);
        }

        std.debug.panic("requirement failed: {s}", .{@errorName(call_error)});
    };
}
