const coral = @import("../coral.zig");

const std = @import("std");

pub const AddressingModel = enum(u32) {
    logical = 0,
};

pub const Capability = enum(u32) {
    matrix = 0,
    shader = 1,
};

pub const Document = struct {
    ids_assigned: u32 = 0,
    ops: coral.Stack(u32) = .empty,

    fn nextId(self: *Document) u32 {
        self.ids_assigned += 1;

        return self.ids_assigned;
    }

    pub fn op(self: *Document, code: u32) std.mem.Allocator.Error!void {
        try self.ops.pushGrow(@byteSwap(code));
    }

    pub fn opCapability(self: *Document, capability: Capability) std.mem.Allocator.Error!void {
        try self.op(17);
        try self.op(@intFromEnum(capability));
    }

    pub fn opMemoryModel(self: *Document, addressing: AddressingModel, memory: MemoryModel) std.mem.Allocator.Error!void {
        try self.op(14);
        try self.op(@intFromEnum(addressing));
        try self.op(@intFromEnum(memory));
    }

    pub fn opExtInstImport(self: *Document, extension: []const u8) std.mem.Allocator.Error!u32 {
        const id = self.nextId();

        try self.op(11);
        try self.op(id);
        try self.string(extension);

        return id;
    }

    pub fn string(self: *Document, data: []const u8) std.mem.Allocator.Error!void {
        const data_len_with_sentinel = data.len + 1;
        const word_count = (data_len_with_sentinel + 4) / 4;

        try self.ops.grow(word_count);

        std.debug.assert(self.ops.pushMany(word_count, 0));

        const buffer = std.mem.sliceAsBytes(self.ops.values[(self.ops.values.len - word_count)..]);

        @memcpy(buffer[0..data.len], data);
    }

    pub fn writeTo(self: Document, spirv: coral.bytes.Writable) coral.bytes.ReadWriteError!void {
        const Header = extern struct {
            magic_number: u32 = 0x07230203, // SPIR-V magic number in little-endian
            version: u32, // SPIR-V version (e.g., 0x00010000 for 1.0)
            generator_magic: u32, // Tool or generator identifier
            id_upper_bound: u32, // Upper bound for IDs
            reserved: u32 = 0, // Reserved, always 0
        };

        try coral.bytes.writeLittle(spirv, Header{
            .magic_number = 0x07230203,
            .version = 0x00010000,
            .generator_magic = 0,
            .id_upper_bound = self.ids_assigned,
        });

        return coral.bytes.writeAll(spirv, std.mem.sliceAsBytes(self.ops.values));
    }
};

pub const MemoryModel = enum(u32) {
    glsl450 = 1,
};

pub fn generate(spirv: coral.bytes.Writable, root: coral.shaders.Root) (coral.shaders.GenerationError || std.mem.Allocator.Error)!void {
    // TODO: Finish and testing coverage.
    var document = Document{};

    try document.opCapability(.shader);

    _ = root;
    _ = try document.opExtInstImport("GLSL.std.450");

    try document.opMemoryModel(.logical, .glsl450);
    try document.writeTo(spirv);
}
