pub const Behavior = @import("./App/Behavior.zig");

pub const Setup = @import("./App/Setup.zig");

const SystemGraph = @import("./App/SystemGraph.zig");

const coral = @import("coral");

const ona = @import("./ona.zig");

const std = @import("std");

initialized_states: coral.map.Hashed(*const coral.TypeId, coral.Box, coral.map.ptrTraits(*const coral.TypeId)),
named_systems: coral.map.Hashed([]const u8, SystemGraph, coral.map.string_traits),
is_running: bool,

pub const RunError = std.mem.Allocator.Error || error{
    MissingDependency,
};

const Self = @This();

pub const Time = struct {
    elapsed: f64,
};

pub fn deinit(self: *Self) void {
    {
        var entries = self.named_systems.entries();

        while (entries.next()) |entry| {
            entry.value.deinit();
        }

        self.named_systems.deinit();
    }

    {
        var entries = self.initialized_states.entries();

        while (entries.next()) |entry| {
            entry.value.deinit();
        }

        self.initialized_states.deinit();
    }
}

pub fn hasState(self: *Self, comptime State: type) ?*State {
    if (self.initialized_states.get(.of(State))) |boxed_state| {
        return boxed_state.has(State).?;
    }

    return null;
}

pub fn init() std.mem.Allocator.Error!Self {
    var self = Self{
        .initialized_states = .empty,
        .named_systems = .empty,
        .is_running = true,
    };

    try self.setState(Time{
        .elapsed = 0,
    });

    return self;
}

pub fn on(self: *Self, comptime schedule: anytype, behavior: *const Behavior) std.mem.Allocator.Error!void {
    try (try self.systemNamed(schedule)).insert(self, behavior);
}

pub fn putState(self: *Self, state: anytype) std.mem.Allocator.Error!*@TypeOf(state) {
    try self.setState(state);

    return self.hasState(@TypeOf(state)).?;
}

pub fn run(self: *Self, tasks: *coral.asio.TaskQueue, comptime schedule: anytype) RunError!void {
    try (try self.systemNamed(schedule)).run(self, tasks);
}

pub fn setState(self: *Self, state: anytype) std.mem.Allocator.Error!void {
    var boxed_state = try coral.Box.init(state);

    errdefer {
        boxed_state.deinit();
    }

    var replaced = try self.initialized_states.replace(.of(@TypeOf(state)), boxed_state);

    if (replaced) |*entry| {
        entry.value.deinit();
    }
}

fn systemNamed(self: *Self, comptime schedule: anytype) std.mem.Allocator.Error!*SystemGraph {
    const schedule_name = switch (@typeInfo(@TypeOf(schedule))) {
        .enum_literal => @tagName(schedule),
        else => @compileError("paramater `schedule` must be an enum literal type"),
    };

    return self.named_systems.get(schedule_name) orelse (try self.named_systems.emplace(schedule_name, .empty)).?;
}
