const SystemGraph = @import("./SystemGraph.zig");

const coral = @import("coral");

const ona = @import("../ona.zig");

const std = @import("std");

label: [*:0]const u8,
on_insertion: *const fn (*const Self, *ona.App, *SystemGraph) std.mem.Allocator.Error!void,
on_run: *const fn (*ona.App) void,
traits: ona.Traits,

const Param = struct {
    init_params: []const InitParam,
    has_bind: ?fn (*ona.App) void = null,
    traits: ona.Traits = .{},

    const InitParam = struct {
        type: type,
        is_required: bool,
        is_read_only: bool,
        traits: ona.Traits,
    };

    fn has_fn(comptime SystemParam: type, comptime name: []const u8) ?std.builtin.Type.Fn {
        switch (@typeInfo(SystemParam)) {
            .@"struct", .@"opaque" => {},
            else => @compileError("system fn params must be either struct or opaque types"),
        }

        if (@hasDecl(SystemParam, name)) {
            return switch (@typeInfo(@TypeOf(@field(SystemParam, name)))) {
                .@"fn" => |@"fn"| @"fn",
                else => @compileError(std.fmt.comptimePrint("system params with a .{s} declaration must be made an fn type", .{name})),
            };
        }

        return null;
    }

    pub fn init(comptime fn_param: std.builtin.Type.Fn.Param) Param {
        if (fn_param.is_generic) {
            @compileError("generic params on behavior fns are disallowed");
        }

        const init_fn = has_fn(fn_param.type.?, "init") orelse {
            @compileError("behavior params must declare a .init fn");
        };

        if (init_fn.return_type != fn_param.type.?) {
            const param_type_name = @typeName(fn_param.type.?);

            @compileError(std.fmt.comptimePrint("{s}.init must return {s}", .{ param_type_name, param_type_name }));
        }

        const init_params = struct {
            const instance = get: {
                var buffer: [init_fn.params.len]InitParam = undefined;

                for (&buffer, init_fn.params) |*init_param, init_fn_param| {
                    if (init_fn_param.is_generic) {
                        @compileError("generic params on Param.init fns are disallowed");
                    }

                    init_param.* = switch (@typeInfo(init_fn_param.type.?)) {
                        .pointer => |pointer| .{
                            .type = pointer.child,
                            .is_required = true,
                            .is_read_only = pointer.is_const,
                            .traits = if (@hasDecl(pointer.child, "traits") and @TypeOf(pointer.child.traits) == ona.Traits) pointer.child.traits else .{},
                        },

                        .optional => |optional| switch (@typeInfo(optional.child)) {
                            .pointer => |pointer| .{
                                .type = pointer.child,
                                .is_required = false,
                                .is_read_only = pointer.is_const,
                            },

                            else => @compileError("non-pointer optional params on Param.init fns are disallowed"),
                        },

                        else => @compileError("params on Param.init fns must be optionals or pointers"),
                    };
                }

                break :get buffer;
            };
        };

        var param = Param{
            .init_params = &init_params.instance,
        };

        if (has_fn(fn_param.type.?, "bind")) |bind_fn| {
            const type_error_message = std.fmt.comptimePrint("{s}.bind must match the signature fn ({s}) !{s}", .{
                @typeName(@TypeOf(fn_param.type.?.bind)),
                @typeName(*ona.App),
                @typeName(void),
            });

            if (bind_fn.params.len != 1) {
                @compileError(type_error_message);
            }

            if (bind_fn.params[0].type.? != *ona.App) {
                @compileError(type_error_message);
            }

            const binding = struct {
                fn bind(app: *ona.App) void {
                    return switch (@typeInfo(bind_fn.return_type.?)) {
                        .error_union => coral.require(fn_param.type.?.bind, .{app}),
                        .void => fn_param.type.?.bind(app),
                        else => @compileError(type_error_message),
                    };
                }
            };

            param.has_bind = binding.bind;
        }

        if (@hasDecl(fn_param.type.?, "traits") and @TypeOf(fn_param.type.?.traits) == ona.Traits) {
            param.traits = fn_param.type.?.traits;
        }

        inline for (param.init_params) |init_param| {
            param.traits = param.traits.derived(init_param.traits);
        }

        return param;
    }
};

const Self = @This();

pub fn after(comptime self: *const Self, comptime dependency: *const Self) *const Self {
    const afters = struct {
        fn on_insertion(behavior: *const Self, app: *ona.App, systems: *SystemGraph) std.mem.Allocator.Error!void {
            try self.on_insertion(behavior, app, systems);
            try systems.depend_on_behavior(app, behavior, dependency);
        }

        fn on_run(app: *ona.App) void {
            self.on_run(app);
        }

        const instance = Self{
            .label = std.fmt.comptimePrint("({s} after {s})", .{ self.label, dependency.label }),
            .on_insertion = on_insertion,
            .on_run = on_run,
        };
    };

    return &afters.instance;
}

pub fn before(comptime self: *const Self, comptime dependant: *const Self) *const Self {
    const afters = struct {
        fn on_insertion(behavior: *const Self, app: *ona.App, systems: *SystemGraph) std.mem.Allocator.Error!void {
            try systems.depend_on_behavior(app, dependant, behavior);
            try systems.insert(app, behavior);
        }

        const instance = Self{
            .label = std.fmt.comptimePrint("({s} before {s})", .{ self.label, dependant.label }),
            .on_insertion = on_insertion,
        };
    };

    return &afters.instance;
}

pub fn of(comptime call: anytype) *const Self {
    const Call = @TypeOf(call);

    const call_fn = switch (@typeInfo(Call)) {
        .@"fn" => |@"fn"| @"fn",
        else => @compileError("`call` parameter must be an fn type"),
    };

    const behaviors = struct {
        fn on_insertion(behavior: *const Self, app: *ona.App, systems: *SystemGraph) std.mem.Allocator.Error!void {
            inline for (call_fn.params) |call_param| {
                const behavior_param = Param.init(call_param);

                inline for (behavior_param.init_params) |init_param| {
                    try systems.dependOnType(app, behavior, .{
                        .id = .of(init_param.type),
                        .is_read_only = init_param.is_read_only,
                    });
                }

                if (behavior_param.has_bind) |bind| {
                    bind(app);
                }
            }
        }

        fn on_run(app: *ona.App) void {
            var call_args: std.meta.ArgsTuple(Call) = undefined;

            inline for (&call_args, call_fn.params) |*call_arg, call_fn_param| {
                const call_param = Param.init(call_fn_param);
                var init_args: std.meta.ArgsTuple(@TypeOf(call_fn_param.type.?.init)) = undefined;

                inline for (&init_args, call_param.init_params) |*init_arg, init_param| {
                    if (init_param.is_required) {
                        init_arg.* = app.hasState(init_param.type) orelse {
                            @panic(std.fmt.comptimePrint("{s} is a required state but not present in the App", .{@typeName(init_param.type)}));
                        };
                    } else {
                        init_arg.* = app.hasState(init_param.type);
                    }
                }

                call_arg.* = @call(.auto, call_fn_param.type.?.init, init_args);
            }

            switch (@typeInfo(call_fn.return_type.?)) {
                .void => @call(.auto, call, call_args),
                .error_union => coral.require(call, call_args),

                else => @compileError(std.fmt.comptimePrint("parameter `call` must be fn (*ona.App) void or fn (*ona.App) !void, not {s}", .{
                    @typeName(Call),
                })),
            }
        }

        const instance = Self{
            .label = @typeName(Call),
            .on_insertion = on_insertion,
            .on_run = on_run,

            .traits = derive: {
                var derived_traits = ona.Traits{};

                for (call_fn.params) |call_param| {
                    derived_traits = derived_traits.derived(Param.init(call_param).traits);
                }

                break :derive derived_traits;
            },
        };
    };

    return &behaviors.instance;
}
