const coral = @import("coral");

const ext = @import("ext");

const ona = @import("./ona.zig");

const std = @import("std");

storage: *ext.SDL_Storage,

pub const Error = error{
    AssetNotFound,
};

const Instance = struct {
    storage: *ext.SDL_Storage,

    pub fn deinit(self: *Instance) void {
        while (!ext.SDL_StorageReady(self.storage)) {
            ext.SDL_Delay(1);
        }

        if (!ext.SDL_CloseStorage(self.storage)) {
            std.log.warn("Failed to close title storage", .{});
        }

        self.storage = undefined;
    }
};

const Self = @This();

pub fn bind(app: *ona.App) !void {
    const title_storage = ext.SDL_OpenTitleStorage(null, 0) orelse {
        std.log.err("Failed to open title storage", .{});

        return error.SdlFailure;
    };

    try app.setState(Instance{
        .storage = title_storage,
    });
}

pub fn init(instance: *Instance) Self {
    return .{
        .storage = instance.storage,
    };
}

pub fn load(self: Self, path: []const u8, allocator: std.mem.Allocator) (std.mem.Allocator.Error || Error)![]u8 {
    const path_buffer = try normalize_path(path);

    while (!ext.SDL_StorageReady(self.storage)) {
        ext.SDL_Delay(1);
    }

    var asset_size: u64 = 0;

    if (!ext.SDL_GetStorageFileSize(self.storage, path_buffer.ptr(), &asset_size)) {
        return error.AssetNotFound;
    }

    const file_buffer = try allocator.alloc(u8, asset_size);

    errdefer {
        allocator.free(file_buffer);
    }

    if (!ext.SDL_ReadStorageFile(self.storage, path_buffer.ptr(), file_buffer.ptr, file_buffer.len)) {
        return error.AssetNotFound;
    }

    return file_buffer;
}

pub fn normalize_path(data: []const u8) Error!coral.ShortString {
    const separator = '/';
    var result = coral.ShortString.empty;

    for (0..data.len) |i| {
        if (data[i] == separator) {
            const next_i = i + 1;

            if (next_i == data.len) {
                continue;
            }

            if (data[next_i] == separator) {
                continue;
            }
        }

        result.append(data[i]) catch |append_error| {
            return switch (append_error) {
                error.StringTooLong => error.AssetNotFound,
            };
        };
    }

    if (result.isEmpty()) {
        return error.AssetNotFound;
    }

    return result;
}

pub fn read(self: Self, path: []const u8, output: []u8) Error!void {
    const path_buffer = try normalize_path(path);

    while (!ext.SDL_StorageReady(self.storage)) {
        ext.SDL_Delay(1);
    }

    if (!ext.SDL_ReadStorageFile(self.storage, path_buffer.ptr(), output.ptr, output.len)) {
        return error.AssetNotFound;
    }
}

pub fn size(self: Self, path: []const u8) Error!usize {
    const path_buffer = try normalize_path(path);
    var asset_size: u64 = 0;

    while (!ext.SDL_StorageReady(self.storage)) {
        ext.SDL_Delay(1);
    }

    if (!ext.SDL_GetStorageFileSize(self.storage, path_buffer.ptr(), &asset_size)) {
        return error.AssetNotFound;
    }

    return asset_size;
}
