const builtin = @import("builtin");

const coral = @import("coral");

const ext = @import("ext");

const ona = @import("./ona.zig");

const std = @import("std");

pub const Display = struct {
    width: u16,
    height: u16,
    is_hidden: bool,
};

const Context = struct {
    window: *ext.SDL_Window,
    gpu_device: *ext.SDL_GPUDevice,
    basic_shader: *ext.SDL_GPUShader = undefined,

    pub fn deinit(self: *Context) void {
        ext.SDL_DestroyGPUDevice(self.gpu_device);
        ext.SDL_DestroyWindow(self.window);
    }

    pub const traits = ona.Traits{
        .is_thread_unsafe = true,
    };
};

fn compile_shaders(_: ona.Write(Context), assets: ona.Assets) !void {
    var arena = std.heap.ArenaAllocator.init(coral.heap.allocator);

    defer {
        arena.deinit();
    }

    const shader_path = "graphics_demo.eff";
    var root = try coral.shaders.Root.init(&arena);

    const Constants = extern struct {
        screen_width: f32,
        screen_height: f32,
        time: f32,
    };

    try root.defineUniform(&arena, .{
        .identifier = "effect",
        .semantic = "Effect",
        .binding = 0,
        .has_field = .fieldsOf(Constants),
    });

    try root.defineTexture(&arena, .{
        .identifier = "albedo",
        .binding = 0,
        .layout = .dimensions_2,
    });

    try root.defineInput(&arena, .{
        .identifier = "uv",
        .type = .float2,
        .location = 0,
    });

    try root.defineOutput(&arena, .{
        .identifier = "color",
        .type = .float4,
        .location = 0,
    });

    switch (try root.parseText(&arena, try assets.load(shader_path, arena.allocator()))) {
        .ok => {
            var codes = coral.Stack(u8).empty;

            try coral.shaders.generateSpirv(coral.bytes.stackWriter(&codes), root);

            std.log.info("{s}", .{codes.values});
        },

        .failure => |failure| {
            std.log.err("failed to parse shader {s}: {s}", .{ shader_path, failure });

            return error.ShaderParseFailure;
        },
    }

    // const basic_shader = ext.SDL_CreateGPUShader(gpu_device, &.{
    //     .code = spirv_code,
    //     .code_size = spirv_code.len,
    //     .stage = ext.SDL_GPU_SHADERSTAGE_FRAGMENT,
    //     .entrypoint = "frag_main",
    // }) orelse {
    //     return sdl_failure("Failed to create basic shader");
    // };
}

pub fn poll(input_events: ona.Send(ona.hid.Event)) !void {
    var event: ext.SDL_Event = undefined;

    while (ext.SDL_PollEvent(&event)) {
        try input_events.push(switch (event.type) {
            ext.SDL_EVENT_QUIT => .quit,
            ext.SDL_EVENT_KEY_UP => .{ .key_up = @enumFromInt(event.key.scancode) },
            ext.SDL_EVENT_KEY_DOWN => .{ .key_down = @enumFromInt(event.key.scancode) },

            ext.SDL_EVENT_MOUSE_BUTTON_UP => .{
                .mouse_up = switch (event.button.button) {
                    ext.SDL_BUTTON_LEFT => .left,
                    ext.SDL_BUTTON_RIGHT => .right,
                    ext.SDL_BUTTON_MIDDLE => .middle,
                    else => continue,
                },
            },

            ext.SDL_EVENT_MOUSE_BUTTON_DOWN => .{
                .mouse_down = switch (event.button.button) {
                    ext.SDL_BUTTON_LEFT => .left,
                    ext.SDL_BUTTON_RIGHT => .right,
                    ext.SDL_BUTTON_MIDDLE => .middle,
                    else => continue,
                },
            },

            ext.SDL_EVENT_MOUSE_MOTION => .{
                .mouse_motion = .{
                    .relative_position = .{ event.motion.xrel, event.motion.yrel },
                    .absolute_position = .{ event.motion.x, event.motion.y },
                },
            },

            else => continue,
        });
    }
}

pub fn prepare(display: ona.Write(Display)) void {
    display.ptr.is_hidden = false;
}

pub fn render(context: ona.Read(Context)) !void {
    const commands = ext.SDL_AcquireGPUCommandBuffer(context.ptr.gpu_device) orelse {
        return error.SdlFailure;
    };

    errdefer {
        _ = !ext.SDL_CancelGPUCommandBuffer(commands);
    }

    var swapchain_texture: *ext.SDL_Texture = undefined;

    if (!ext.SDL_WaitAndAcquireGPUSwapchainTexture(commands, context.ptr.window, @ptrCast(&swapchain_texture), null, null)) {
        return error.SdlFailure;
    }

    {
        const color_targets = [_]ext.SDL_GPUColorTargetInfo{
            .{
                .texture = @ptrCast(swapchain_texture),
                .load_op = ext.SDL_GPU_LOADOP_CLEAR,
                .clear_color = .{ .r = 0, .g = 0.2, .b = 0.4, .a = 1.0 },
            },
        };

        const render_pass = ext.SDL_BeginGPURenderPass(commands, &color_targets, color_targets.len, null) orelse {
            return error.SdlFailure;
        };

        defer {
            ext.SDL_EndGPURenderPass(render_pass);
        }
    }

    if (!ext.SDL_SubmitGPUCommandBuffer(commands)) {
        return;
    }
}

pub fn setup(app: *ona.App) !void {
    if (!ext.SDL_Init(ext.SDL_INIT_VIDEO)) {
        return error.SystemFailure;
    }

    const display = try app.putState(Display{
        .width = 1280,
        .height = 720,
        .is_hidden = false,
    });

    const window = ext.SDL_CreateWindow("Ona", display.width, display.height, ext.SDL_WINDOW_HIDDEN) orelse {
        return error.SdlFailure;
    };

    errdefer {
        ext.SDL_DestroyWindow(window);
    }

    const gpu_device = ext.SDL_CreateGPUDevice(ext.SDL_GPU_SHADERFORMAT_SPIRV, builtin.mode == .Debug, null) orelse {
        return error.SdlFailure;
    };

    errdefer {
        ext.SDL_DestroyGPUDevice(gpu_device);
    }

    if (!ext.SDL_ClaimWindowForGPUDevice(gpu_device, window)) {
        return error.SdlFailure;
    }

    try app.setState(Context{
        .window = window,
        .gpu_device = gpu_device,
    });

    try app.on(.load, .of(prepare));
    try app.on(.load, .of(compile_shaders));
    try app.on(.post_update, .of(poll));
    try app.on(.render, .of(render));
    try app.on(.finish, .of(synchronize));
}

pub fn synchronize(context: ona.Write(Context), display: ona.Read(Display)) !void {
    const window_flags = ext.SDL_GetWindowFlags(context.ptr.window);
    const is_window_hidden = (window_flags & ext.SDL_WINDOW_HIDDEN) != 0;

    if (is_window_hidden != display.ptr.is_hidden) {
        const was_window_changed = switch (display.ptr.is_hidden) {
            true => ext.SDL_HideWindow(context.ptr.window),
            false => ext.SDL_ShowWindow(context.ptr.window),
        };

        if (!was_window_changed) {
            std.log.warn("failed to change window visibility", .{});
        }
    }
}
